"""@package docstring
B-ASIC Signal Module.
"""
from typing import Optional, TYPE_CHECKING

from b_asic.graph_component import TypeName
from b_asic.abstract_graph_component import AbstractGraphComponent
from b_asic.graph_component import Name

if TYPE_CHECKING:
    from b_asic.port import InputPort, OutputPort

class Signal(AbstractGraphComponent):
    """A connection between two ports."""

    _source: "OutputPort"
    _destination: "InputPort"

    def __init__(self, source: Optional["OutputPort"] = None, \
                destination: Optional["InputPort"] = None, name: Name = ""):

        super().__init__(name)

        self._source = source
        self._destination = destination

        if source is not None:
            self.connect_source(source)

        if destination is not None:
            self.connect_destination(destination)

    @property
    def source(self) -> "OutputPort":
        """Returns the source OutputPort of the signal."""
        return self._source

    @property
    def destination(self) -> "InputPort":
        """Returns the destination "InputPort" of the signal."""
        return self._destination

    def connect_source(self, src: "OutputPort") -> None:
        """Disconnects the previous source OutputPort of the signal and
        connects to the entered source OutputPort. Also connects the entered
        source port to the signal if it hasn't already been connected.

        Keyword arguments:
        - src: OutputPort to connect as source to the signal.
        """
        self.disconnect_source()
        self._source = src
        if not src.is_connected_to_signal(self):
            # If the new source isn't connected to this signal then connect it.
            src.connect_signal(self)

    def connect_destination(self, dest: "InputPort") -> None:
        """Disconnects the previous destination InputPort of the signal and
        connects to the entered destination InputPort. Also connects the entered
        destination port to the signal if it hasn't already been connected.

        Keywords argments:
        - dest: InputPort to connect as destination to the signal.
        """
        self.disconnect_destination()
        self._destination = dest
        if not dest.is_connected_to_signal(self):
            # If the new destination isn't connected to tis signal then connect it.
            dest.connect_signal(self)

    @property
    def type_name(self) -> TypeName:
        return "s"

    def disconnect_source(self) -> None:
        """Disconnects the source OutputPort of the signal. If the source port
        still is connected to this signal then also disconnects the source port."""
        if self.has_source():
            old_source: "OutputPort" = self._source
            self._source = None
            if old_source.is_connected_to_signal(self):
                # If the old destination port still is connected to this signal, then disconnect it.
                old_source.disconnect_signal_by_ref(self)

    def disconnect_destination(self) -> None:
        """Disconnects the destination "InputPort" of the signal."""
        if self.has_destination():
            old_destination: "InputPort" = self._destination
            self._destination = None
            if old_destination.is_connected_to_signal(self):
                # If the old destination port still is connected to this signal, then disconnect it.
                old_destination.disconnect_signal()

    def port_is_source(self, source: "OutputPort") -> bool:
        """Returns true if the entered OutputPort is the source else false."""
        return self._source is source

    def port_is_destination(self, destination: "InputPort") -> bool:
        """Returns true if the entered OutputPort is the destination else false."""
        return self._destination is destination

    def is_connected(self) -> bool:
        """Returns true if the signal is connected to both a source and a destination,
        else false."""
        return self.has_source() and self.has_destination()

    def has_source(self) -> bool:
        """Returns true if the signal is connected to a source, else false."""
        return self._source is not None

    def has_destination(self) -> bool:
        """Returns true if the signal is connected to a destination, else false."""
        return self._destination is not None
